;;; <pattern matching.scm>
; Nach Robert Sedgewick 'Algorithmen' S.336
; 'Algorithmus von Boyer und Moore'
; Gesucht werden Zeichenketten in Zeichenketten
; Sein Beispiel : "sting" in
; "a string searching example consisting of ..."
; 1. Untersuche zu jedem mglichen Zeichen des Textes, wie
;    weit bei dem vorgegebenen Muster zu springen wre,
;    wenn keine bereinstimmung gefunden wird.
; 2. Whle das letzte Zeichen des Musters und das an der 
;    zugehrigen Position im Text. 
; 3. Wenn sie nicht uebereinstimmen, springe um die Zahl von
;    Positionen weiter, die in der Liste (s.1.) abgelegt ist.
; 4. Wenn sie uebereinstimmen, vergleiche das ganze Muster.
;    { Bei hufig auftretenden bereinstimmungen wre es hier
;    noch mglich eine Verbesserung zu erzielen. }
; 5. Stimmt das Muster nicht berein, springe um die Muster-
;    lnge weiter.

;;; ===== Sprung =============================================
; Zu 1.:
; Die Prozedur erzeugt die Assoziationsliste der Sprungweiten.
; Zu jedem mglichen Zeichen des Textes wird die Sprungweite
; abhngig vom Auftreten des Zeichens im Muster berechnet.
; In der Prozedur wird das Muster zunchst in eine Liste ver-
; wandelt, da Zugriffe dann einfacher sind.
; member kann als Prdikat und als Zugriffsfunktion verwendet
; werden.

(define
  (sprung text-zeichen muster)
 (let
     loop
   ((text-zeichen text-zeichen)
    (muster (string->list muster)))
   (cond
     ((null? text-zeichen) ()) ; rekursiv nachklappernd
     ((member (car text-zeichen) muster)
      (cons             ; Teilliste z.B.: (#\i 3) =3.Position
       (list 
        (car text-zeichen)
        (sub1 (length (member (car text-zeichen) muster))))
       (loop (cdr text-zeichen) muster)))
     (else
      (cons
       (list (car text-zeichen) (length muster))
       (loop (cdr text-zeichen) muster))))))

;;; ===== Filter =============================================
; Bei sehr langen Texten sollte text-zeichen aus allen Zeichen
; bestehen, sonst werden mit Filter alle nicht vorkommenden
; Zeichen entfernt.
(define
  (filter text)
  (let
      loop
    ((code 20)
     (liste ())
     (text (string->list text)))
    (cond
      ((> code 255) (reverse liste))
      ((member (integer->char code) text)
       (loop (add1 code) (cons (integer->char code) liste) text))
      (else
       (loop (add1 code) liste text)))))
;;; ===== match ==============================================
; holt sich die Assoziationsliste der Sprungweiten mit 'sprung'
; Dann wird an der Position der letzten Musterbuchstaben im
; Text begonnen zu berprfen, ob er mit dem Textzeichen ber-
; einstimmt. Wenn nicht, wird entsprechend der Sprungtabelle
; weitergesprungen.
; Stimmt er berein, wird hier nach einem einfachen Verfahren
; einfach berprft, ob das ganze Muster mit der Zeichenfolge
; im Text bereinstimmt. Dies knnte noch verbessert werden.
; Wieder unbedingt zu beachten sind die Zugriffsindizes fr
; strings: 0..Laenge-1 , daher einige add1 und sub1 !
; Extern wird die Position aber 'normal' von 1..Laenge
; gezaehlt.
(define
  (match text muster)
  (let
      ((asso (sprung (filter text) muster))
       (laenge (string-length muster))
       (text-laenge (string-length text)))
    (let
        loop
      ((position (sub1 laenge)))
      (cond
        ((>= position text-laenge) #f)
        ((char=? 
          (string-ref muster (sub1 laenge))
          (string-ref text position))
         (if
          (string=?
           muster
           (substring 
            text
            (- (add1 position) laenge) 
            (add1 position)))
          (begin
            (display "Gefunden: ")
            (display muster)
            (display " an Position ")
            (display (add1 (- (add1 position) laenge)))
            (newline)
            (add1 position))
          (loop
           (add1 position))))
        (else
         (loop 
          (+ position
             (cadr
              (assoc 
               (string-ref text position)
               asso)))))))))

;;; ===== Test von filter ====================================
;(filter "a string searching example consisting of ...")
;;; ===== Test von sprung ====================================
;(sprung 
; (filter "a string searching example consisting of ...") 
; "sting")
;;; ===== Test von match =====================================
(match 
 "a string searching example consisting of ..."
 "a s")
(match 
 "a string searching example consisting of ..."
 "sting")
(match 
 "a string searching example consisting of ..."
 "amp")
(match 
 "a string searching example consisting of ..."
 "nsisti")
(match 
    (string-append
 "alle voegel sind schon da, alle voegel alle. "
 "amsel, drossel, fink und star und die ganze vogelschaar, "
 "wuenschen dir ein frohes jahr, "
 "lauter heil und segen. ")
 "star")
